# Idempotent producer
The idempotent producer strengthens Kafka's delivery semantics from at least once to exactly once delivery.
In particular producer retries will no longer introduce duplicates




# Transactional producer
The transactional producer allows an application to send messages to multiple partitions (and topics!) atomically

The purpose of the `transactionalId` is to enable transaction recovery across multiple sessions of a single producer instance.
It would typically be derived from the shard identifier in a partitioned, stateful, application.
As such, it should be unique to each producer instance running within a partitioned application.

There can be only one open transaction per producer

```java
Properties props = new Properties();
props.put("bootstrap.servers", "localhost:9092");
props.put("transactional.id", "my-transactional-id");
Producer<String, String> producer = new KafkaProducer<>(props, new StringSerializer(), new StringSerializer());

producer.initTransactions();

try {
   producer.beginTransaction();
   for (int i = 0; i < 100; i++) {
     producer.send(new ProducerRecord<>("my-topic", Integer.toString(i), Integer.toString(i)));
   }
   producer.commitTransaction();
} catch (ProducerFencedException | OutOfOrderSequenceException | AuthorizationException e) {
   // We can't recover from these exceptions, so our only option is to close the producer and exit.
   producer.close();
} catch (KafkaException e) {
   // For all other exceptions, just abort the transaction and try again.
   producer.abortTransaction();
}

producer.close();
```

__NOTE__: TransactionManager class to maintain state for transactions

## initTransactions

1. Ensures any transactions initiated by previous instances of the producer with the same
   transactional.id are completed. If the previous instance had failed with a transaction in
   progress, it will be aborted. If the last transaction had begun completion,
   but not yet finished, this method awaits its completion.
2. Gets the internal producer id and epoch, used in all future transactional
   messages issued by the producer.

## beginTransaction

`ProducerFencedException` if another producer with the same `transactional.id` is active

In memory operation, signals to the TransactionManager the state IN_TRANSACTION

## commitTransaction

Set TransactionManager state to COMMITTING_TRANSACTION

- Calls AddPartitionsToTxn
- Calls EndTxnRequest with COMMIT

## abortTransaction

- Clear newPartitionsInTransaction
- Calls EndTxnRequest with ABORT

## __NEW__ tx.add

- maybeAddPartitionToTransaction
  - Set<> newPartitionsInTransaction
- add record to RecordAccumulator

## Transactional Producer API

```javascript
const producer = kafka.producer({ transactional: true })

const run = async () => {
  // should call initProducerId
  // new log message to highlight the use of a transactional producer
  await producer.connect()

  const tx = await producer.tx()

  try {
    for (let i = 0; i < 100; i++) {
      await tx.add({
        topic,
        messages: [{ key: `key-${i}`, value: `value-${i}` }]
      })
    }
    await tx.commit()
  } catch (e) {
    // Investigate class of errors where the producer can't recover, i.e: ProducerFencedException
    await tx.abort()
  }

  // regular cluster disconnect
  await producer.disconnect()
}
```
